"""Time series feature generator functions."""
from typing import Dict, List, Optional, Union

import holidays
import numpy as np
import pandas as pd

from tsfeast.utils import to_list


def get_busdays_in_month(dt: pd.Timestamp) -> int:
    """
    Get the number of business days in a month period, using US holidays.

    Parameters
    ----------
    dt: pd.Timestamp
        Desired month.

    Returns
    -------
    int
        Number of business days in the month.
    """
    chooser: Dict[bool, pd.Timestamp] = {
        True: dt,
        False: dt - pd.tseries.offsets.MonthBegin()
    }
    month_begin = chooser[dt.is_month_start]
    month_end = dt + pd.tseries.offsets.MonthBegin(1)  # np.busday_count end date is exclusive
    us_holidays = list(holidays.US(years=dt.year).keys())
    return np.busday_count(month_begin.date(), month_end.date(), holidays=us_holidays)  # type: ignore #pylint: disable=line-too-long # noqa


def get_datetime_features(
        data: Union[pd.DataFrame, pd.Series], date_col: Optional[str] = None,
        dt_format: Optional[str] = None, freq: Optional[str] = None
) -> pd.DataFrame:
    """
    Get features based on datetime index, including year, month, week, weekday, quarter, days in
    month, business days in month and leap year.

    Parameters
    ----------
    data: pd.DataFrame, pd.Series
        Original data.
    date_col: Optional[str]
        Column name containing date/timestamp.
    dt_format: Optional[str]
        Date/timestamp format, e.g. `%Y-%m-%d` for `2020-01-31`.
    freq: Optional[str]
        Date frequency.

    Returns
    -------
    pd.DataFrame
        Date features.
    """
    if isinstance(data, pd.DataFrame):
        if date_col is None:
            raise ValueError('`date_col` cannot be none when passing a DataFrame.')
        dates = data[date_col]
    elif isinstance(data, pd.Series):
        dates = data
    else:
        raise ValueError('`data` must be a DataFrame or Series.')

    if not freq:
        freq = pd.infer_freq(pd.DatetimeIndex(pd.to_datetime(dates, format=dt_format)))

    X_dt = pd.DatetimeIndex(pd.to_datetime(dates, format=dt_format))  # enforce DatetimeIndex
    dt_features = pd.DataFrame()
    dt_features['year'] = X_dt.year  # pylint: disable=no-member
    dt_features['quarter'] = X_dt.quarter  # pylint: disable=no-member
    dt_features['month'] = X_dt.month  # pylint: disable=no-member
    if freq == 'D':
        dt_features['week'] = X_dt.week  # pylint: disable=no-member
        dt_features['weekday'] = X_dt.weekday  # pylint: disable=no-member
    if freq and 'M' in freq:
        dt_features['days_in_month'] = X_dt.days_in_month  # pylint: disable=no-member
        dt_features['bdays_in_month'] = pd.Series(X_dt).apply(get_busdays_in_month)
    dt_features['leap_year'] = X_dt.is_leap_year.astype(int)  # pylint: disable=no-member
    dt_features.index = data.index
    return dt_features


def get_lag_features(data: pd.DataFrame, n_lags: int) -> pd.DataFrame:
    """
    Get n-lagged features for data.

    Parameters
    ----------
    data: pd.DataFrame
        Original data.
    n_lags: int
        Number of lags to generate.

    Returns
    -------
    pd.DataFrame
        Lagged values of specified dataset.

    """
    lags = []
    for n in range(1, n_lags+1):
        df = data.copy().shift(n)
        df.columns = [f'{x}_lag_{n}' for x in data.columns]
        lags.append(df)
    return pd.concat(lags, axis=1)


def get_rolling_features(data: pd.DataFrame, window_lengths: List[int]) -> pd.DataFrame:
    """
    Get rolling metrics (mean, std, min, max) for each specified window length.

    Parameters
    ----------
    data: pd.DataFrame
        Original data.
    window_lengths: List[int]
        List of window lengths to generate.

    Returns
    -------
    pd.DataFrame
        Rolling mean, std, min and max for each specified window length.

    """
    window_lengths = to_list(window_lengths)
    df = data.copy()
    windows = []
    metrics = ['sum', 'mean', 'std', 'min', 'max']
    for win in window_lengths:
        for m in metrics:
            windows.append(
                pd.DataFrame(
                    df.rolling(win).agg(m).values,
                    columns=[f'{c}_{win}_pd_{m}' for c in df.columns],
                    index=df.index
                )
            )
    return pd.concat(windows, axis=1)


def get_ewma_features(data: pd.DataFrame, window_lengths: List[int]) -> pd.DataFrame:
    """
    Get an exponentially-weighted moving average for each specified window length.

    Parameters
    ----------
    data: pd.DataFrame
        Original data.
    window_lengths: List[int]
        List of window lengths to generate.

    Returns
    -------
    pd.DataFrame
        Exponentially-weighted moving average for each specified window length.

    """
    window_lengths = to_list(window_lengths)
    df = data.copy()
    windows = []
    for win in window_lengths:
        windows.append(
            pd.DataFrame(
                df.ewm(span=win, min_periods=win-1).mean().values,
                columns=[f'{c}_{win}_ewm_mean' for c in df.columns],
                index=df.index
            )
        )
    return pd.concat(windows, axis=1)


def get_change_features(data: pd.DataFrame, period_lengths: List[int]) -> pd.DataFrame:
    """
    Get percent change for all features for each specified period length.

    Parameters
    ----------
    data: pd.DataFrame
        Original data.
    period_lengths: List[int]
        A list of period lengths to generate.

    Returns
    -------
    pd.DataFrame
        Percent changes for all features.
    """
    period_lengths = to_list(period_lengths)
    df = data.copy()
    changes = []
    for period in period_lengths:
        changes.append(
            pd.DataFrame(
                df.pct_change(periods=period).values,
                columns=[f'{c}_{period}_pd_pct_chg' for c in df.columns],
                index=df.index
            )
        )
    return pd.concat(changes, axis=1)


def get_difference_features(data: pd.DataFrame, n_diffs: int) -> pd.DataFrame:
    """
    Get n differences for all features.

    Parameters
    ----------
    data: pd.DataFrame
        Original data.
    n_diffs: int
        Number of differences to return.

    Returns
    -------
    pd.DataFrame
        N-differenced data.
    """
    diffs = []
    for n in range(1, n_diffs + 1):
        df = data.copy().shift(n)
        df.columns = [f'{x}_diff_{n}' for x in data.columns]
        diffs.append(df)
    return pd.concat(diffs, axis=1)



import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline

from tests.conftest import train_test
from tsfeast.tsfeatures import TimeSeriesFeatures


class TestTimeSeriesFeatures:
    def test_fit_constant(self, exog):
        ts = TimeSeriesFeatures(datetime='index', trend='c')
        actual = ts.fit_transform(exog.reset_index())
        assert actual['const'].sum() == float(len(exog))

    def test_fit_trend(self, exog):
        ts = TimeSeriesFeatures(datetime='index', trend='ct')
        actual = ts.fit_transform(exog.reset_index())['trend']
        expected = list(range(1, len(exog)+1))
        np.testing.assert_allclose(actual, expected)

    def test_fit_datetime_only(self, exog, curr_dir):
        ts = TimeSeriesFeatures(datetime='index', interactions=False)
        actual = ts.fit_transform(exog.reset_index())
        fp = curr_dir.joinpath('valid_outputs', 'dt_transform_only.json')
        expected = pd.read_json(fp).reset_index(drop=True)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)

    def test_fit_all(self, exog, curr_dir):
        ts = TimeSeriesFeatures(
            datetime='index', trend='ct', lags=4, rolling=[3, 6], ewma=[3, 6], pct_chg=[1, 12],
            diffs=1, polynomial=2, interactions=True
        )
        actual = ts.fit_transform(exog.reset_index())
        fp = curr_dir.joinpath('valid_outputs', 'all_transforms.json')
        expected = pd.read_json(fp).fillna(0).reset_index(drop=True)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False, check_dtype=False)

    def test_pipeline(self, curr_dir):
        x_train, y_train, x_test, y_test = train_test()
        ts = TimeSeriesFeatures(
            datetime='index', trend='ct', lags=4, rolling=[3, 6], ewma=[3, 6], pct_chg=[1, 12],
            diffs=1, polynomial=2, interactions=True
        )
        pl = Pipeline([
            ('features', ts),
            ('regression', LinearRegression())
        ])
        pl.fit(x_train.reset_index(), y_train)
        pl.predict(x_test.reset_index())
        actual = pl.named_steps.features.output_features_
        fp = curr_dir.joinpath('valid_outputs', 'all_transforms.json')
        expected = pd.read_json(fp).fillna(0).iloc[-len(x_test):, :].astype(float)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)



import pathlib

import numpy as np
import pandas as pd
import pytest

HERE = pathlib.Path(__file__).parent
N = 12
IDX = pd.date_range(start='2020-01', periods=N, freq='M')
y = pd.Series(np.arange(100, 100+N), index=IDX, name='series-0')
Y = pd.DataFrame(np.arange(100, 100+(N*3)).reshape(N, 3), index=IDX, columns=[f'series-{x}' for x in range(3)])
X = pd.DataFrame(np.arange(10, 10+(N*5)).reshape(N, 5), index=IDX, columns=[f'feature-{x}' for x in range(5)])
dates = [x.strftime('%Y-%m-%d') for x in IDX]


@pytest.fixture
def curr_dir():
    return HERE


@pytest.fixture
def endog_uni():
    return y


@pytest.fixture
def endog_multi():
    return Y


@pytest.fixture
def exog():
    return X


@pytest.fixture()
def exog_dt():
    return X.reset_index()


@pytest.fixture
def date_col():
    return dates


def train_test():
    x_train, x_test = X.iloc[:10, :], X.iloc[-2:, :]
    y_train, y_test = y.iloc[:10], y.iloc[-2]
    return x_train, y_train, x_test, y_test



from unittest.mock import MagicMock, Mock

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from tests.conftest import HERE, X, train_test
from tsfeast.funcs import (
    get_change_features,
    get_datetime_features,
    get_difference_features,
    get_ewma_features,
    get_lag_features,
    get_rolling_features,
)
from tsfeast.transformers import (
    BaseTransformer,
    ChangeFeatures,
    DateTimeFeatures,
    DifferenceFeatures,
    EwmaFeatures,
    InteractionFeatures,
    LagFeatures,
    OriginalFeatures,
    PolyFeatures,
    RollingFeatures,
    Scaler,
)

x_train, y_train, x_test, y_test = train_test()


class TestBaseTransformer:
    def test_array(self, monkeypatch, exog):
        t = OriginalFeatures()
        mock = MagicMock(pd.DataFrame)
        monkeypatch.setattr('tsfeast.transformers.array_to_dataframe', mock)
        t.transform(exog.values)
        mock.assert_called()

    def test_transform_raises(self, exog):
        with pytest.raises(NotImplementedError):
            t = BaseTransformer()
            t._transform(exog)


def test_inverse_scaler(exog):
    sc = Scaler()
    trans = sc.fit_transform(exog)
    actual = sc.inverse_transform(trans)
    expected = X.astype(float)
    pd.testing.assert_frame_equal(actual, expected)


def test_bad_data_raises():
    dt = DateTimeFeatures('index')
    with pytest.raises(ValueError):
        dt.fit([[1, 2, 3], [4, 5, 6]])


class TestOriginalFeatures:
    @pytest.mark.parametrize(
        'input, transformer, expected',
        [
            (X, OriginalFeatures(), X),
            (X.reset_index(), DateTimeFeatures('index'), get_datetime_features(X.reset_index(), 'index')),
            (X, LagFeatures(2), get_lag_features(X, n_lags=2).fillna(0)),
            (X, LagFeatures(2, fillna=False), get_lag_features(X, n_lags=2)),
            (X, RollingFeatures([4]), get_rolling_features(X, [4]).fillna(0)),
            (X, EwmaFeatures([4]), get_ewma_features(X, [4]).fillna(0)),
            (X, ChangeFeatures([4]), get_change_features(X, [4]).fillna(0)),
            (X, DifferenceFeatures(2), get_difference_features(X, 2).fillna(0)),
            (X, PolyFeatures(3), pd.read_json(HERE.joinpath('valid_outputs', 'poly_[2, 3].json'))),
            (X, InteractionFeatures(), pd.read_json(HERE.joinpath('valid_outputs', 'interactions.json')).astype(float)),
            (X, Scaler(), pd.DataFrame(StandardScaler().fit_transform(X), columns=X.columns, index=X.index))
        ]
    )
    def test_solo(self, endog_uni, input, transformer, expected):
        feat = transformer
        actual = feat.fit_transform(input, endog_uni)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)

    @pytest.mark.parametrize(
        'transformer, expected',
        [
            (OriginalFeatures(), x_test),
            (DateTimeFeatures('index'), get_datetime_features(x_test.reset_index(), 'index', freq='M')),
            (LagFeatures(2), get_lag_features(X, n_lags=2).fillna(0).iloc[-len(x_test):, :]),
            (RollingFeatures([4]), get_rolling_features(X, [4]).fillna(0).iloc[-len(x_test):, :]),
            (EwmaFeatures([4]), get_ewma_features(X, [4]).fillna(0).iloc[-len(x_test):, :]),
            (ChangeFeatures([4]), get_change_features(X, [4]).fillna(0).iloc[-len(x_test):, :]),
            (DifferenceFeatures(2), get_difference_features(X, 2).fillna(0).iloc[-len(x_test):, :]),
            (PolyFeatures(3), pd.read_json(HERE.joinpath('valid_outputs', 'poly_[2, 3].json')).iloc[-len(x_test):, :]),
            (InteractionFeatures(), pd.read_json(HERE.joinpath('valid_outputs', 'interactions.json')).iloc[-len(x_test):, :].astype(float))
        ]
    )
    def test_pipeline(self, transformer, expected):
        pl = Pipeline([
            ('features', transformer),
            ('regression', LinearRegression())
        ])
        if isinstance(transformer, DateTimeFeatures):
            pl.fit(x_train.reset_index(), y_train)
            pl.predict(x_test.reset_index())
        else:
            pl.fit(x_train, y_train)
            pl.predict(x_test)
        actual = pl.named_steps.features.output_features_
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)



import pytest

from tsfeast.funcs import *


class TestGetBusdaysInMonth:
    def test_dt_monthbegin(self):
        actual = get_busdays_in_month(pd.to_datetime('2020-01-01'))
        expected = 21
        assert actual == expected

    def test_dt_monthend(self):
        actual = get_busdays_in_month(pd.to_datetime('2020-01-31'))
        expected = 21
        assert actual == expected

    @pytest.mark.parametrize(
        "test_date, result", [
            (pd.to_datetime('2020-01-15'), 21),
            (pd.to_datetime('2020-02-15'), 19),
            (pd.to_datetime('2020-04-15'), 22),
        ]
    )
    def test_dt_other(self, test_date, result):
        actual = get_busdays_in_month(pd.to_datetime(test_date))
        expected = result
        assert actual == expected


class TestGetDatetimeFeatures:
    def test_no_date_col_raises(self, exog, curr_dir):
        with pytest.raises(ValueError):
            actual = get_datetime_features(exog.reset_index())

    def test_bad_type_raise(self, exog):
        with pytest.raises(ValueError):
            actual = get_datetime_features(exog.reset_index().to_dict())

    def test_outputs(self, exog, curr_dir):
        actual = get_datetime_features(exog.reset_index(), date_col='index')
        fp = curr_dir.joinpath('valid_outputs', 'uni_dt_features.json')
        expected = pd.read_json(fp).reset_index(drop=True)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)


class TestGetLagFeatures:
    @pytest.mark.parametrize(
        "lag", [1, 3, 6]
    )
    def test_multivariate(self, exog, lag, curr_dir):
        actual = get_lag_features(exog, lag)
        fp = curr_dir.joinpath('valid_outputs', f'multi_lags_{lag}.json')
        expected = pd.read_json(fp)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)


class TestGetRollingFeatures:
    @pytest.mark.parametrize(
        "win", [[2], [3, 6]]
    )
    def test_multivariate(self, exog, win, curr_dir):
        actual = get_rolling_features(exog, win)
        fp = curr_dir.joinpath('valid_outputs', f'multi_rolling_{win}.json')
        expected = pd.read_json(fp)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)

    @pytest.mark.parametrize(
        "i, expected", [
            (0, 135.0),
            (1, 141.0),
            (6, 23.5),
            (10, 9.35414),
            (15, 10.0),
            (20, 35.0)
        ]
    )
    def test_metrics(self, exog, i, expected):
        actual = get_rolling_features(exog, [6]).iloc[5, i]
        assert round(actual, 5) == expected


class TestGetEwmFeatures:
    @pytest.mark.parametrize(
        "win", [[2], [3, 6]]
    )
    def test_multivariate(self, exog, win, curr_dir):
        actual = get_ewma_features(exog, win)
        fp = curr_dir.joinpath('valid_outputs', f'multi_ewm_{win}.json')
        expected = pd.read_json(fp)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)


class TestGetChangeFeatures:
    @pytest.mark.parametrize(
        "win", [1, 6]
    )
    def test_multivariate(self, exog, win, curr_dir):
        actual = get_change_features(exog, win)
        fp = curr_dir.joinpath('valid_outputs', f'multi_pct_chg_{win}.json')
        expected = pd.read_json(fp)
        pd.testing.assert_frame_equal(actual, expected, check_freq=False)


class TestGetDifferenceFeatures:
    def test_multivariate(self, exog, curr_dir):
        actual = get_difference_features(exog, 2)
        fp = curr_dir.joinpath('valid_outputs', f'differences.json')
        expected = pd.read_json(fp)
        np.testing.assert_allclose(actual, expected)



import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_regression
from sklearn.linear_model import LinearRegression

from tsfeast.utils import array_to_dataframe, array_to_series, plot_diag

ARR = np.array([
            [99.99999518, 100.99999518, 101.99999518, 102.99999518],
            [103.99999518, 104.99999518, 105.99999518, 106.99999518],
            [107.99999518, 108.99999518, 109.99999518, 110.99999518]
        ])

X, y = make_regression(n_features=25)


def test_array_to_dataframe():
    actual = array_to_dataframe(ARR)
    expected = pd.DataFrame(ARR, columns=['x0', 'x1', 'x2', 'x3'])
    pd.testing.assert_frame_equal(actual, expected)


def test_array_to_series():
    actual = array_to_series(np.array([0, 1, 2, 3, 4, 5]))
    expected = pd.Series(actual, name='y')
    pd.testing.assert_series_equal(actual, expected)


class TestPlotDiag:
    def test_both_none_raises(self):
        with pytest.raises(ValueError):
            plot_diag(residuals=None, estimator=None)

    def test_no_x_raises(self):
        with pytest.raises(ValueError):
            plot_diag(estimator=LinearRegression, X=None, y=ARR)

    def test_no_y_raises(self):
        with pytest.raises(ValueError):
            plot_diag(estimator=LinearRegression, X=ARR, y=None)

    def test_estimator(self):
        lr = LinearRegression()
        lr.fit(X, y)
        plot_diag(estimator=lr, X=X, y=y)

    def test_resid(self):
        lr = LinearRegression()
        lr.fit(X, y)
        resid = y - lr.predict(X)
        plot_diag(resid)



"""Miscellaneous utility functions."""
from typing import List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.linear_model._base import LinearModel
from statsmodels.graphics.gofplots import qqplot
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf

Data = Union[pd.DataFrame, pd.Series, np.ndarray]


def to_list(x: Union[int, List]) -> List[int]:
    """Ensure parameter is list of integer(s)."""
    if isinstance(x, list):
        return x
    return [x]


def array_to_dataframe(x: np.ndarray) -> pd.DataFrame:
    """Convert Numpy array to Pandas DataFrame with default column names."""
    return pd.DataFrame(x, columns=[f'x{i}' for i in range(x.shape[1])])


def array_to_series(x: np.ndarray) -> pd.Series:
    """Convert Numpy array to Pandas Series with default name."""
    return pd.Series(x, name='y')


def plot_diag(
        residuals: Optional[Data] = None, estimator: Optional[LinearModel] = None,
        X: Optional[Data] = None, y: Optional[Data] = None
):
    """
    Plot regression diagnostics.

    Generate residuals plot, QQ plot, ACF plot and PACF plot, given either an array-like object
    of residuals or and estimator and X and y data arrays.

    Parameters
    ----------
    residuals: Data
        Model residual errors.
    estimator: LinearModel
        Scikit-Learn generalized linear model.
    X: array of shape [n_samples, n_features]
            The input samples.
    y :  array-like of shape (n_samples,) or (n_samples, n_outputs), default=None
        Target values (None for unsupervised transformations).

    Raises
    ------
    ValueError
        - If neither residuals or estimator provider.
        - If estimator provided without X and y data.
    """
    if residuals is None and estimator is None:
        raise ValueError('Either residuals or estimator and X, y must be given.')
    if estimator and (X is None or y is None):
        raise ValueError('Both X and y must be given if passing an estimator.')
    if residuals is None and y is not None and estimator is not None:
        residuals = y - estimator.predict(X)
    _, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 10))
    ax1.plot(residuals)
    ax1.set_title('Residuals')
    qqplot(residuals, line='s', ax=ax2)
    ax2.set_title('QQ-Plot')
    plot_acf(residuals, ax=ax3)
    plot_pacf(residuals, ax=ax4)



import numpy as np
import pandas as pd
import pytest

from tsfeast.splitter import EndogSeriesWindows, TimeSeriesWindows


class TestTimeSeriesWindows:
    def test_split_sets(self, endog_uni, exog):
        tsw = TimeSeriesWindows(train_length=3, test_length=1, gap_length=0)
        windows = tsw.split(endog_uni, exog)
        for w in windows:
            assert len(w) == 4

    @pytest.mark.parametrize(
        'train_length, test_length', [
            (2, 1),
            (3, 1),
            (3, 2)
        ]
    )
    def test_num_splits(self, endog_uni, exog, train_length, test_length):
        tsw = TimeSeriesWindows(train_length=train_length, test_length=test_length, gap_length=0)
        windows = tsw.split(endog_uni, exog)
        assert len(windows) == (len(endog_uni) - train_length - (test_length - 1))

    @pytest.mark.parametrize(
        'train_length, test_length, gap_length', [
            (2, 1, 1),
            (3, 1, 2),
            (3, 2, 1)
        ]
    )
    def test_gap_n_splits(self, endog_uni, exog, train_length, test_length, gap_length):
        tsw = TimeSeriesWindows(train_length=train_length, test_length=test_length, gap_length=gap_length)
        windows = tsw.split(endog_uni, exog)
        assert len(windows) == (len(endog_uni) - (train_length + gap_length) - (test_length - 1))

    @pytest.mark.parametrize(
        'train_length, test_length, gap_length', [
            (2, 1, 1),
            (3, 1, 2),
            (3, 2, 0)
        ]
    )
    def test_split_shape(self, endog_uni, exog, train_length, test_length, gap_length):
        tsw = TimeSeriesWindows(train_length=train_length, test_length=test_length, gap_length=gap_length)
        windows = tsw.split(endog_uni, exog)
        for split in windows:
            x_train, x_test, y_train, y_test = split
            assert x_train.shape[0] == train_length
            assert x_test.shape[0] == test_length
            assert y_train.shape[0] == train_length
            assert y_test.shape[0] == test_length

    @pytest.mark.parametrize(
        'train_length, test_length, gap_length', [
            (2, 1, 1),
            (3, 1, 2),
            (3, 2, 0)
        ]
    )
    def test_gap_exists(self, endog_uni, exog, train_length, test_length, gap_length):
        tsw = TimeSeriesWindows(train_length=train_length, test_length=test_length, gap_length=gap_length)
        windows = tsw.split(endog_uni, exog)
        for split in windows:
            x_train, x_test, y_train, y_test = split
            assert (x_train.index[-1] + pd.tseries.offsets.MonthEnd(1+gap_length)) == x_test.index[0]
            assert (y_train.index[-1] + pd.tseries.offsets.MonthEnd(1 + gap_length)) == y_test.index[0]



import pytest

from tsfeast._base import BaseContainer


class TestBaseEstimator:
    def test_fit_not_implemented(self, exog, endog_uni):
        mod = BaseContainer()
        with pytest.raises(NotImplementedError):
            mod._fit(exog, endog_uni)

    def test_predict_not_implemented(self, exog, endog_uni):
        mod = BaseContainer()
        with pytest.raises(NotImplementedError):
            mod._predict(exog)



"""
Time Series Windows Module.

*Note* These classes split data into n, equal-length, sliding training/test windows.  This differs
from Scikit-Learn's TimeSeriesSplit implementation where windows are accumulated:

Scikit-Learn
-------------
Win 0   |----|
Win 1   |--------|
Win2    |------------|

TimeSeriesWindows
-----------------
Win 0   |----|--------
Win 1   -|----|-------
Win2    --|----|------
"""
from typing import List, Optional

import pandas as pd

# pylint: disable=missing-docstring


class TimeSeriesWindows:
    def __init__(self, train_length: int, test_length: int, gap_length: int = 0) -> None:
        self.train_length = train_length
        self.test_length = test_length
        self.gap_length = gap_length

    def split(self, y: pd.DataFrame, x: pd.DataFrame) -> List[pd.DataFrame]:
        windows = []
        for i in range(len(x)):
            train_start = i
            train_end = i + self.train_length
            test_start = train_end + self.gap_length
            test_end = test_start + self.test_length
            if test_end <= len(x):
                x_train = x.iloc[train_start:train_end]
                y_train = y.iloc[train_start:train_end]
                x_test = x.iloc[test_start:test_end]
                y_test = y.iloc[test_start:test_end]
                split = x_train, x_test, y_train, y_test
                windows.append(split)
        return windows


class EndogSeriesWindows(TimeSeriesWindows):
    def __init__(
            self, min_train_length: int, test_length: int, max_train_length: Optional[int] = None,
            gap_length: int = 0) -> None:
        super().__init__(min_train_length, test_length, gap_length)
        self.min_train_length = min_train_length
        self.max_train_length = max_train_length

    def split(self, y: pd.DataFrame, x=None) -> List[pd.DataFrame]:
        windows = []
        for i in range(self.min_train_length, len(y)):
            if i + self.test_length <= len(y):
                if self.max_train_length is not None:
                    train_start = i - self.max_train_length
                else:
                    train_start = 0
                train_end = i
                if self.gap_length is not None:
                    test_start = train_end + self.gap_length
                else:
                    test_start = train_end
                test_end = test_start + self.test_length
                y_train = y.iloc[train_start:train_end]
                y_test = y.iloc[test_start:test_end]
                split = y_train, y_test
                windows.append(split)
        return windows



"""Custom scoring metrics."""
from numpy import log
from sklearn.metrics import mean_squared_error

from tsfeast._base import BaseContainer
from tsfeast.utils import Data


def bic_score(mse: float, n: int, p: int):
    """
    Calcuate BIC score.

    Parameters
    ----------
    mse: float
        Mean-squared error.
    n: int
        Number of observations.
    p: int
        Number of parameters

    Returns
    -------
    float
        BIC value.
    """
    return n * log(mse) + log(n) * p


def bic_scorer(estimator: BaseContainer, X: Data, y: Data):
    """Score SciKit-Learn estimator using BIC."""
    y_pred = estimator.predict(X)
    mse = mean_squared_error(y, y_pred)
    return bic_score(mse, X.shape[0], X.shape[1])



"""Module for Base Estimator."""
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

from tsfeast.utils import Data


class BaseContainer(BaseEstimator, RegressorMixin):
    """Container class for Scikit-Learn models."""
    def __init__(self):
        """Instantiate container."""

    def _fit(self, X: Data, y: Data) -> Data:
        """Method not implemented."""
        raise NotImplementedError

    def fit(self, X: Data, y: Data) -> "BaseContainer":
        """
        Fit the estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.
        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), default=None
            Target values (None for unsupervised transformations).
        Returns
        -------
        BaseContainer
            Self.
        """
        X, y = check_X_y(X, y)
        self._fit(X, y)
        return self

    def _predict(self, X: Data) -> Data:
        """Method not implemented."""
        raise NotImplementedError

    def predict(self, X: Data) -> Data:
        """
        Make predictions with fitted estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        Returns
        -------
        np.ndarray
            Array of predicted values.
        """
        check_is_fitted(self)
        X = check_array(X)
        return self._predict(X)



"""
Module for Scikit-Learn Regressor with ARMA Residuals and
Scikit-Learn API wrapper for Statsmodels TSA models.
"""
from typing import Tuple

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.linear_model._base import LinearModel
from statsmodels.base.model import Model
from statsmodels.tsa.arima.model import ARIMA

from tsfeast._base import BaseContainer
from tsfeast.utils import Data


class ARMARegressor(BaseContainer):
    """
    Estimator for Scikit-Learn estimator with ARMA residuals.

    Parameters
    ----------
    estimator: LinearModel
        Scikit-Learn linear estimator.
    order: Tuple[int, int, int]
        ARIMA order for residuals.

    Attributes
    ----------
    estimator: LinearModel
        The Scikit-Learn regressor.
    order: Tuple[int, int, int]
        The (p,d,q,) order of the ARMA model.
    intercept_: float
        The fitted estimator's intercept.
    coef_: np.ndarray
        The fitted estimator's coefficients.
    arma_: ARIMA
        The fitted ARMA model.
    fitted_values_: np.ndarray
        The combined estimator and ARMA fitted values.
    resid_: np.ndarray
        The combined estimator and ARMA residual values.
    """
    def __init__(
            self, estimator: LinearModel = LinearRegression(),
            order: Tuple[int, int, int] = (1, 0, 0)
    ):
        """
        Instantiate ARMARegressor object.

        Parameters
        ----------
        estimator: LinearRegression
            Scikit-Learn linear estimator.
        order: Tuple[int, int, int]
            ARIMA order for residuals.

        """
        super().__init__()
        self.estimator = estimator
        self.order = order

    def _fit(self, X: Data, y: Data) -> "ARMARegressor":
        """
        Fit the estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.
        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), default=None
            Target values (None for unsupervised transformations).
        Returns
        -------
        ARMARegressor
            Self.
        """
        self.estimator.fit(X, y)
        self.intercept_ = self.estimator.intercept_
        self.coef_ = self.estimator.coef_
        estimator_fitted = self.estimator.predict(X)
        estimator_resid = np.ravel(y) - estimator_fitted
        self.arma_ = ARIMA(estimator_resid, order=self.order).fit()
        self.fitted_values_ = estimator_fitted + self.arma_.fittedvalues
        self.resid_ = np.ravel(y) - self.fitted_values_
        return self

    def _predict(self, X: Data) -> Data:
        """
        Predict the response.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        Returns
        -------
        np.ndarray
            Array of predicted values.

        Notes
        -----
        The `predict` method should not be used to get fitted values from the training set; rather,
        users should access this same data using the `fitted_values_` attribute.  The `predict`
        method calls the ARMA regresor's forecast method, which generates predictions from the last
        time step in the training data, thus does not match, temporally, with the training data.
        """
        estimator_pred = self.estimator.predict(X)
        # todo this won't work for in-sample predictions
        arma_pred = self.arma_.forecast(steps=estimator_pred.shape[0])
        return estimator_pred + arma_pred


class TSARegressor(BaseContainer):
    """
    Estimator for StatsModels TSA model.

    Parameters
    ----------
    model: Model
        An uninstantiated Statsmodels TSA model.
    use_exog: bool
        Whether to use exogenous features; default False.
    kwargs:
        Additional kwargs for Statsmodels model.

    Attributes
    ----------
    fitted_model_: Model
        The fitted Statmodels model object.
    summary_:
        The fitted Statmodels model summary results.

    """
    def __init__(self, model: Model, use_exog: bool = False, **kwargs):
        """
        Instantiate TSARegressor object.

        model: Model
            An uninstantiated Statsmodels TSA model.
        use_exog: bool
            Whether to use exogenous features; default False.
        kwargs:
            Additional kwargs for Statsmodels model.
        """
        super().__init__()
        self.model = model
        self.use_exog = use_exog
        self.kwargs = kwargs

    def _fit(self, X: Data, y: Data) -> "TSARegressor":
        """
        Fit the estimator.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.
        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), default=None
            Target values (None for unsupervised transformations).
        Returns
        -------
        TSARegressor
            Self.
        """
        if self.use_exog:
            self.fitted_model_ = self.model(endog=y, exog=X, **self.kwargs).fit()
        else:
            self.fitted_model_ = self.model(endog=y, **self.kwargs).fit()
        self.fitted_values_ = self.fitted_model_.fittedvalues
        self.summary_ = self.fitted_model_.summary()
        return self

    def _predict(self, X: Data) -> Data:
        """
        Predict the response.

        Parameters
        ----------
        X : array of shape [n_samples, n_features]
            The input samples.

        Returns
        -------
        np.ndarray
            Array of predicted values.
        """
        return self.fitted_model_.forecast(steps=X.shape[0])



import pytest
from sklearn.linear_model import LinearRegression

from tsfeast.metrics import bic_score, bic_scorer


def test_bic_score():
    actual = bic_score(100, 12, 5)
    expected = 67.6865754807971
    assert round(actual, 4) == round(expected, 4)


def test_bic_scorer(monkeypatch, exog, endog_uni):
    mod = LinearRegression()
    monkeypatch.setattr(mod, 'predict', lambda *args: endog_uni+10)
    actual = bic_scorer(mod, exog, endog_uni)
    expected = 67.6865754807971
    assert round(actual, 4) == round(expected, 4)



"""Time series feature generators as Scikit-Learn compatible transformers."""

from itertools import combinations
from typing import List, Optional

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import PolynomialFeatures, StandardScaler
from sklearn.utils.validation import check_is_fitted

from tsfeast.funcs import (
    get_change_features,
    get_datetime_features,
    get_difference_features,
    get_ewma_features,
    get_lag_features,
    get_rolling_features,
)
from tsfeast.utils import Data, array_to_dataframe


class BaseTransformer(BaseEstimator, TransformerMixin):
    """Base transformer object."""

    def __init__(self, fillna: bool = True):
        """Instantiate transformer object."""
        self.fillna = fillna

    def transform(self, X: Data, y=None) -> Data:
        """
        Transform fitted data.

        Parameters
        ----------
        X: array of shape [n_samples, n_features]
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Array-like object of transformed data.

        Notes
        -----
        Scikit-Learn Pipelines only call the `.transform()` method during the `.predict()` method,
        which is appropriate to prevent data leakage in predictions.  However, most of the
        transformers in this module take a set of features and generate new features; there's no
        inherent method to transform some timeseries features given a fitted estimator.

        For time series lags, changes, etc., we have access to past data for feature
        generation without risk of data leakage; certain features (e.g. lags) require this
        to avoid NaNs or zeros.

        We append new X to our original features and transform on entire dataset, keeping
        only the last n rows.  Appropriate for time series transformations, only.
        """
        if isinstance(X, np.ndarray):
            X = array_to_dataframe(X)

        if hasattr(self, 'input_features_'):
            rows = X.shape[0]
            X = pd.concat([self.input_features_, X])  # pylint: disable=E0203
            self.output_features_ = self._transform(X, y).iloc[-rows:, :]
            if self.fillna:
                return self.output_features_.fillna(0)
            return self.output_features_
        self.input_features_: pd.DataFrame = X
        self.n_features_in_ = X.shape[0]
        self.output_features_ = self._transform(X, y)
        self.feature_names_ = self.output_features_.columns
        if self.fillna:
            return self.output_features_.fillna(0)
        return self.output_features_

    def get_feature_names(self) -> List[str]:
        """Get list of feature names."""
        check_is_fitted(self)
        return list(self.feature_names_)

    def _transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform input data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        raise NotImplementedError

    def fit(self, X: Data, y=None) -> "BaseTransformer":
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: array of shape [n_samples, n_features]
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        BaseTransformer
            Self.
        """
        _, _ = X, y
        return self


class OriginalFeatures(BaseTransformer):
    """Return original features."""

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return X


class Scaler(BaseTransformer):
    """Wrap StandardScaler to maintain column names."""

    def __init__(self):
        """Instantiate transformer object."""
        super().__init__()
        self.scaler = StandardScaler()

    def fit(self, X: pd.DataFrame, y=None) -> "Scaler":
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        self.scaler.fit(X)
        return self

    def _transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        return self

    def transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        self.feature_names_ = X.columns
        return pd.DataFrame(
            self.scaler.transform(X),
            columns=X.columns,
            index=X.index
        )

    def inverse_transform(self, X: pd.DataFrame, copy: bool = True) -> pd.DataFrame:
        """
        Transform scaled data into original feature space.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        copy: bool
            Default True; if False, try to avoid a copy and do inplace scaling instead.

        Returns
        -------
        Data
            Data in original feature space.
        """
        return pd.DataFrame(
            self.scaler.inverse_transform(X, copy=copy),
            columns=self.feature_names_,
            index=X.index
        )


class DateTimeFeatures(BaseTransformer):
    """Generate datetime features."""

    def __init__(
            self, date_col: Optional[str] = None, dt_format: Optional[str] = None,
            freq: Optional[str] = None
    ):
        """
        Instantiate transformer object.

        date_col: Optional[str]
            Column name containing date/timestamp.
        dt_format: Optional[str]
            Date/timestamp format, e.g. `%Y-%m-%d` for `2020-01-31`.
        """
        super().__init__()
        self.date_col = date_col
        self.dt_format = dt_format
        self.freq = freq

    def fit(self, X: Data, y=None) -> "DateTimeFeatures":
        _ = y
        if isinstance(X, pd.DataFrame):
            dates = X[self.date_col]
        elif isinstance(X, pd.Series):
            dates = X
        else:
            raise ValueError('`data` must be a DataFrame or Series.')
        if not self.freq:
            self.freq = pd.infer_freq(
                pd.DatetimeIndex(pd.to_datetime(dates, format=self.dt_format))
            )
        return self

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_datetime_features(X, self.date_col, dt_format=self.dt_format, freq=self.freq)


class LagFeatures(BaseTransformer):
    """Generate lag features."""

    def __init__(self, n_lags: int, fillna: bool = True):
        """
        Instantiate transformer object.

        Parameters
        ----------
        n_lags: int
            Number of lags to generate.
        """
        super().__init__(fillna=fillna)
        self.n_lags = n_lags

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_lag_features(X, n_lags=self.n_lags)


class RollingFeatures(BaseTransformer):
    """Generate rolling features."""

    def __init__(self, window_lengths: List[int], fillna: bool = True):
        """
        Instantiate transformer object.

        Parameters
        ----------
        window_lengths: L:ist[int]
            Length of window(s) to create.
        """
        super().__init__(fillna=fillna)
        self.window_lengths = window_lengths

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_rolling_features(X, window_lengths=self.window_lengths)


class EwmaFeatures(BaseTransformer):
    """Generate exponentially-weighted moving-average features."""

    def __init__(self, window_lengths: List[int], fillna: bool = True):
        """
        Instantiate transformer object.

        Parameters
        ----------
        window_lengths: L:ist[int]
            Length of window(s) to create.
        """
        super().__init__(fillna=fillna)
        self.window_lengths = window_lengths

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_ewma_features(X, window_lengths=self.window_lengths)


class ChangeFeatures(BaseTransformer):
    """Generate period change features."""

    def __init__(self, period_lengths: List[int], fillna: bool = True):
        """
        Instantiate transformer object.

        Parameters
        ----------
        period_lengths: List[int]
            Length of period[s] to generate change features.
        """
        super().__init__(fillna=fillna)
        self.period_lengths = period_lengths

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_change_features(X, period_lengths=self.period_lengths)


class DifferenceFeatures(BaseTransformer):
    """Generate difference features."""

    def __init__(self, n_diffs: int, fillna: bool = True):
        """
        Instantiate transformer object.

        Parameters
        ----------
        n_diffs: int
            Number of differences to calculate.
        """
        super().__init__(fillna=fillna)
        self.n_diffs = n_diffs

    def _transform(self, X: pd.DataFrame, y=None) -> Data:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        return get_difference_features(X, n_diffs=self.n_diffs)


class PolyFeatures(BaseTransformer):
    """Generate polynomial features."""

    def __init__(self, degree=2):
        """
        Instantiate transformer object.

        Parameters
        ----------
        degree: int
            Degree of polynomial to use.
        """
        super().__init__()
        self.degree = degree

    def _transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        poly = []
        df = X.copy()
        for i in range(2, self.degree + 1):
            poly.append(
                pd.DataFrame(
                    df.values ** i,
                    columns=[f'{c}^{i}' for c in df.columns],
                    index=df.index
                )
            )
        return pd.concat(poly, axis=1)


class InteractionFeatures(BaseTransformer):
    """Wrap PolynomialFeatures to extract interactions and keep column names."""

    def _transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Fit transformer object to data.

        Parameters
        ----------
        X: pd.DataFrame
            The input samples.
        y: None
            Not used; included for compatibility, only.

        Returns
        -------
        Data
            Transformed features.
        """
        transformer = PolynomialFeatures(interaction_only=True, include_bias=False)
        interactions = transformer.fit_transform(X.fillna(0))
        cols = [':'.join(x) for x in combinations(X.columns, r=2)]
        return pd.DataFrame(
            interactions[:, X.shape[1]:],  # drop original values
            columns=cols,
            index=X.index
        )



"""Time series features module."""
from typing import List, Optional

import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import FeatureUnion
from statsmodels.tsa.tsatools import add_trend

from tsfeast.transformers import (
    BaseTransformer,
    ChangeFeatures,
    DateTimeFeatures,
    DifferenceFeatures,
    EwmaFeatures,
    InteractionFeatures,
    LagFeatures,
    OriginalFeatures,
    PolyFeatures,
    RollingFeatures,
)


class TimeSeriesFeatures(BaseTransformer):
    """Generate multiple time series feature in one transformer."""

    def __init__(
            self, datetime: str, trend: str = 'n', lags: Optional[int] = None,
            rolling: Optional[List[int]] = None, ewma: Optional[List[int]] = None,
            pct_chg: Optional[List[int]] = None, diffs: Optional[int] = None,
            polynomial: Optional[int] = None, interactions: bool = True,
            fillna: bool = True
    ):
        """Instanatiate transformer object."""
        super().__init__()
        self.datetime = datetime
        self.trend = trend
        self.lags = lags
        self.rolling = rolling
        self.ewma = ewma
        self.pct_chg = pct_chg
        self.diffs = diffs
        self.polynomial = polynomial
        self.interactions = interactions
        self.fillna = fillna

    def _transform(self, X, y=None):
        """Fit transformer to data."""
        if not hasattr(self, 'freq_'):
            self.freq_ = pd.infer_freq(pd.DatetimeIndex(pd.to_datetime(X[self.datetime])))

        transforms = {
            'lags': LagFeatures(self.lags),
            'rolling': RollingFeatures(self.rolling),
            'ewma': EwmaFeatures(self.ewma),
            'pct_chg': ChangeFeatures(self.pct_chg),
            'diffs': DifferenceFeatures(self.diffs),
            'polynomial': PolyFeatures(self.polynomial),
            'interactions': InteractionFeatures()
        }
        # don't want `_` attributes from BaseTransformer.fit() method
        self.steps_ = [k for k, v in vars(self).items() if '_' not in k and v]
        numeric = X.select_dtypes('number').columns
        try:
            union = FeatureUnion([(k, v) for k, v in transforms.items() if k in self.steps_])
            if len(union.transformer_list) > 0:
                transformer = ColumnTransformer([
                    ('original', OriginalFeatures(), numeric),
                    ('datetime', DateTimeFeatures(freq=self.freq_), self.datetime),
                    ('features', union, numeric)
                ])
            else:
                transformer = ColumnTransformer([
                    ('original', OriginalFeatures(), numeric),
                    ('datetime', DateTimeFeatures(freq=self.freq_), self.datetime),
                ])
        except ValueError:
            transformer = ColumnTransformer([
                ('original', OriginalFeatures(), numeric),
                ('datetime', DateTimeFeatures(freq=self.freq_), self.datetime)
            ])

        features = pd.DataFrame(
            transformer.fit_transform(X, y), columns=transformer.get_feature_names()
        )
        if self.trend:
            features = add_trend(features, trend=self.trend, prepend=True, has_constant='add')
        if self.fillna:
            return features.fillna(0)
        return features



import numpy as np
import pandas as pd
import pytest
from statsmodels.iolib.summary import Summary
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from statsmodels.tsa.statespace.sarimax import SARIMAX

from tsfeast.models import ARMARegressor, TSARegressor

RTOL = 1e-3
VALID = {
    'arma': {
        'fit': np.array([
            99.99999518, 100.99999518, 101.99999518, 102.99999518,
            103.99999518, 104.99999518, 105.99999518, 106.99999518,
            107.99999518, 108.99999518, 109.99999518, 110.99999518
        ]),
        'resid': np.array([
            4.82455816e-06, 4.82455816e-06, 4.82455816e-06, 4.82455816e-06,
            4.82455816e-06, 4.82455816e-06, 4.82455816e-06, 4.82455816e-06,
            4.82455816e-06, 4.82455816e-06, 4.82455816e-06, 4.82455816e-06
        ]),
        'int': 97.6,
        'coef': np.array([0.04, 0.04, 0.04, 0.04, 0.04]),
        'arma_fit': np.array([
            -4.82455817e-06, -4.82455817e-06, -4.82455817e-06, -4.82455817e-06,
            -4.82455817e-06, -4.82455817e-06, -4.82455817e-06, -4.82455817e-06,
            -4.82455817e-06, -4.82455817e-06, -4.82455817e-06, -4.82455817e-06
        ])
    },
    'tsa': {
        'fit': np.array([
            100.00000001, 100., 100.99999999, 101.99999999,
            102.99999999, 103.99999999, 104.99999999, 105.99999999,
            106.99999999, 107.99999999, 108.99999999, 109.99999999
        ]),
        'forecast': np.array([
            110.99999999, 110.99999999, 110.99999999, 110.99999999,
            110.99999999, 110.99999999, 110.99999999, 110.99999999,
            110.99999999, 110.99999999, 110.99999999, 110.99999999
        ]),
        'with_kwargs': np.array([
            99.99969881, 100.99939108, 101.99987387, 103.00021847,
            104.00043141, 105.0005278, 106.00052128, 107.00042408,
            108.00024716, 109.00000034, 109.99969241, 110.99933123
        ]),
        'with_exog': np.array([
            100., 101., 102., 103., 104., 105., 106., 107., 108., 109., 110.,
            111.
        ])
    },
}


class TestArmaRegressor:
    @pytest.mark.parametrize(
        'attr, expected',
        [
            ('fitted_values_', VALID['arma']['fit']),
            ('intercept_', VALID['arma']['int']),
            ('coef_', VALID['arma']['coef']),
            ('resid_', VALID['arma']['resid']),
        ]
    )
    def test_fit(self, exog, endog_uni, attr, expected):
        mod = ARMARegressor()
        mod.fit(exog, endog_uni)
        actual = getattr(mod, attr)
        np.testing.assert_allclose(actual, expected)

    def test_arma_fit(self, exog, endog_uni):
        mod = ARMARegressor()
        mod.fit(exog, endog_uni)
        actual = mod.arma_.fittedvalues
        expected = VALID['arma']['arma_fit']
        np.testing.assert_allclose(actual, expected)

    def test_predict(self, exog, endog_uni):
        mod = ARMARegressor()
        mod.fit(exog, endog_uni)
        actual = mod.predict(exog)
        np.testing.assert_allclose(actual, VALID['arma']['fit'])


class TestTsaRegressor:
    def test_fit_no_kwargs(self, exog, endog_uni):
        mod = TSARegressor(model=ExponentialSmoothing)
        mod.fit(exog, endog_uni)
        actual = mod.fitted_values_
        expected = VALID['tsa']['fit']
        np.testing.assert_allclose(actual, expected, rtol=RTOL)

    def test_fit_kwargs(self, exog, endog_uni):
        mod = TSARegressor(
            model=ExponentialSmoothing,
            damped=True,
            trend='mul'
        )
        mod.fit(exog, endog_uni)
        actual = mod.fitted_values_
        expected = VALID['tsa']['with_kwargs']
        np.testing.assert_allclose(actual, expected, rtol=RTOL)

    def test_with_exog(self, exog, endog_uni):
        mod = TSARegressor(
            model=SARIMAX,
            use_exog=True,
            order=(1, 0, 0),
            trend='c'
        )
        mod.fit(exog, endog_uni)
        actual = mod.fitted_values_
        expected = VALID['tsa']['with_exog']
        np.testing.assert_allclose(actual, expected)

    def test_predict(self, exog, endog_uni):
        mod = TSARegressor(model=ExponentialSmoothing)
        mod.fit(exog, endog_uni)
        actual = mod.predict(exog)
        expected = VALID['tsa']['forecast']
        np.testing.assert_allclose(actual, expected)

    def test_summary(self, exog, endog_uni):
        mod = TSARegressor(model=ExponentialSmoothing)
        mod.fit(exog, endog_uni)
        assert isinstance(mod.summary_, Summary)
